﻿using Kivii.HumanResources.DingTalk.Entities;
using Kivii.Web;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Kivii.HumanResources.DingTalk.Transforms
{
    public class DingTalkDeptQuery : RestfulExecution<DingTalkDept>
    {
        public string ThirdConfigInternalCode { get; set; }

        /// <summary>
        /// 父部门ID。如果不传，默认部门为根部门，根部门ID为1。只支持查询下一级子部门，不支持查询多级子部门。
        /// </summary>
        public long DeptId { get; set; }

        /// <summary>
        /// 通讯录语言： zh_CN（默认）：中文 en_US：英文
        /// </summary>
        public string Language { get; set; }

        public override object OnExecution(IRequest req, IResponse res)
        {
            ThirdConfigInternalCode.ThrowIfNullOrEmpty("请传入ThirdConfigInternalCode");

            if (Language.IsNullOrEmpty()) Language = "zh_CN";
            if (DeptId <= 0) DeptId = 1;

            var accessToken = DingTalkExtension.GetAccessToken(ThirdConfigInternalCode);

            var result = DingTalkExtension.GetDingTalkDept(accessToken, DeptId, Language);

            var rtns = new RestfulQueryResponse<DingTalkDept>();
            rtns.Results = new List<DingTalkDept>();
            rtns.Results.AddRange(result);
            rtns.Total = rtns.Results.Count;
            return rtns;
        }
    }

    public class DingTalkAttendanceQuery : RestfulExecution<DingTalkAttendance>
    {
        public string ThirdConfigInternalCode { get; set; }

        /// <summary>
        /// 查询考勤打卡记录的起始工作日。格式为“yyyy-MM-dd HH:mm:ss”，HH:mm:ss可以使用00:00:00，将返回此日期从0点到24点的结果。起始与结束工作日最多相隔7天。
        /// </summary>
        public DateTime BeginTime { get; set; }

        /// <summary>
        /// 查询考勤打卡记录的结束工作日。格式为“yyyy-MM-dd HH:mm:ss”，HH:mm:ss可以使用00:00:00，将返回此日期从0点到24点的结果。起始与结束工作日最多相隔7天。
        /// </summary>
        public DateTime EndTime { get; set; }

        /// <summary>
        /// 部门ID
        /// </summary>
        public List<long> DeptIds { get; set; }

        /// <summary>
        /// 表示获取考勤数据的起始点。第一次传0，如果还有多余数据，下次获取传的offset值为之前的offset+limit，012依次递增。
        /// </summary>
        public int Offset { get; set; }

        /// <summary>
        /// 表示获取考勤数据的条数，最大不能超过50条。
        /// </summary>
        public int Limit { get; set; }

        /// <summary>
        /// 是否为海外企业使用：true：海外平台使用 false（默认）：国内平台使用
        /// </summary>
        public bool IsL18n { get; set; }

        public override object OnExecution(IRequest req, IResponse res)
        {
            ThirdConfigInternalCode.ThrowIfNullOrEmpty("请传入ThirdConfigInternalCode");

            if (Limit <= 0) Limit = 20;
            int monthDay = DateTime.DaysInMonth(DateTime.Now.Year, DateTime.Now.Month);
            if (EndTime == DateTime.MinValue) EndTime = new DateTime(DateTime.Now.Year, DateTime.Now.Month, DateTime.Now.Day);
            if (BeginTime == DateTime.MinValue) BeginTime = EndTime.AddDays(-7);

            BeginTime = DateTime.Parse(BeginTime.ToString("yyyy-MM-dd"));
            EndTime = DateTime.Parse(EndTime.ToString("yyyy-MM-dd"));
            if (EndTime < BeginTime) throw new Exception("查询结束日期不可小于开始日期!");

            var accessToken = DingTalkExtension.GetAccessToken(ThirdConfigInternalCode);

            var deptUsers = new List<DingTalkDeptUser>();
            foreach(var deptId in DeptIds)
            {
                var results = DingTalkExtension.GetDingTalkDeptUser(accessToken, deptId, "", Offset, Limit);
                deptUsers.AddRange(results);
            }

            var userIds = deptUsers.ConvertAll(o => o.Userid);
            userIds.ThrowIfNullOrEmpty("所选部门不存在用户,请重新选择!");

            var attendances = DingTalkExtension.GetDingTalkAttendanceList(accessToken, BeginTime, EndTime, userIds, Offset, Limit);
            foreach(var item in attendances)
            {
                var deptUser = deptUsers.FirstOrDefault(o => o.Userid == item.UserId);
                if (deptUser == null) continue;
                item.UserName = deptUser.Name;
            }
            var rtns = new RestfulQueryResponse<DingTalkAttendance>();
            rtns.Results = new List<DingTalkAttendance>();
            rtns.Results.AddRange(attendances);
            rtns.Total = rtns.Results.Count;
            return rtns;
        }
    }

}
